mod mkpsxiso;
use cdtypes::{Error, types::helper::{convert_ascii_to_str, force_convert_ascii_to_str}};
use fast_xml::{events::{Event, attributes::{AttrError, Attribute, Attributes}}, Reader};
use core::ops::Deref;

pub use mkpsxiso::parse as parse_mkpsxiso;

#[derive(Clone)]
struct XMLReader<'a> {
    reader: Reader<&'a [u8]>,
    buffer: Vec<u8>,
}

impl<'a> XMLReader<'a> {
    pub fn new(xml: &str) -> XMLReader {
        let mut reader = Reader::from_str(xml.as_ref());
    
        reader.trim_text(true);
        reader.expand_empty_elements(true);
        reader.check_end_names(true);

        XMLReader{reader, buffer: Vec::new()}
    }

    pub fn next(&mut self) -> Result<Event, Error> {
        self.buffer.clear();
        Self::convert_error(self.reader.read_event(&mut self.buffer))
    }

    pub fn get_attribute(attribute: Result<Attribute, AttrError>) -> Result<Attribute, Error> {
        match attribute {
            Ok(attribute) => Ok(attribute),
            Err(error)    => Err(Error::GenericError(error.to_string())),
        }
    }

    pub fn get_attribute_value_str<'b>(attribute: &'b Attribute) -> Result<&'b str, Error> {
        if let Some(str) = convert_ascii_to_str(attribute.value.deref()) {
            Ok(str)
        }
    
        else {
            Err(Error::GenericError("Invalid UTF-8 for attribute value".to_string()))
        }
    }

    pub fn get_attribute_key_str<'b>(attribute: &'b Attribute) -> Result<&'b str, Error> {
        if let Some(str) = convert_ascii_to_str(attribute.key) {
            Ok(str)
        }
    
        else {
            Err(Error::GenericError("Invalid UTF-8 for attribute key".to_string()))
        }
    }

    pub fn find_attribute<'b>(attributes: Attributes<'b>, name: &[u8]) -> Result<Option<Attribute<'b>>, Error> {
        for attribute in attributes {
            match attribute {
                Ok(attribute) => (
                    if attribute.key == name {
                        return Ok(Some(attribute));
                    }
                ),
    
                Err(error)    => {
                    return Err(Error::GenericError(error.to_string()));
                }
            }
        }
    
        Ok(None)
    }

    pub fn for_each_attribute<'b, F>(attributes: Attributes<'b>, mut callback: F) -> Result<(), Error> where F: FnMut(Attribute) -> Result<(), Error> {
        for attribute in attributes {
            callback(Self::get_attribute(attribute)?)?;
        }

        Ok(())
    }

    pub fn print_attribute_not_supported(name: &[u8], tag_name: &str) {
        println!("Attribute \"{}\" not yet supported for tag \"{}\"", force_convert_ascii_to_str(name), tag_name);
    }

    fn convert_error(result: Result<Event, fast_xml::Error>) -> Result<Event, Error> {
        match result {
            Ok(result) => Ok(result),
            Err(error) => Err(Error::GenericError(error.to_string())),
        }
    }
}